//
//  Bucket
//
//  Bucket is based on the Internet meme with a walrus having a bucket
//  and losing it shortly. The walrus then laments the loss of his bucket.
//
//  If you have any questions regarding the code, contact me
//  via the method specified at  http://www.sebastianmihai.com
//  (operational in 2012), and I will try my best to help.
//
//                             Sebastian Mihai, 2012
//

#include <string.h>
#include "snesc/snesc.h"
#include "tiles.h"
#include "palette.h"

#define OUTLIT(x) (*((char*)0xfffe) = (x))
void write(char* s) {
  while(*s) OUTLIT(*s++);
}

#define MAX_TILES 14

#define BEAR_SPEED 1
#define MAX_WALRUSES 10
#define MIN_WALRUS_SPEED 1

#define BUCKET_ON_GROUND 0
#define BUCKET_HELD 1

int walrusX[ MAX_WALRUSES ];
int walrusY[ MAX_WALRUSES ];
int walrusDX[ MAX_WALRUSES ];
char walrusVisible[ MAX_WALRUSES ];

// we build a background map and a palette inside these
unsigned int blockmap[0x400];
unsigned int pal[0x100];

// sprites
//
// [0][1]
// [2][3]   Bear
//
// [4] bucket
//
// [5][6] walrus
// [7][8] walrus
// [9][10] walrus
// [11][12] walrus
//

unsigned int NextRandom = 1337;

unsigned int random() {
  NextRandom = ((NextRandom * 14577) % 190) + 25;
  return NextRandom;
} 

void SetBackgroundTile( unsigned int x, unsigned int y, unsigned int tile ) {
	blockmap[ 32*y + x ] = tile;
}

void RenderWalrus( unsigned char which ) {

	unsigned int flags = 0x31;
	unsigned char leftTile = 9, rightTile = 10;
	unsigned char temp;
	
	if( walrusVisible[which] == 0) {
		leftTile = 11;
		rightTile = 11;
	}
	
	if( walrusDX[ which ] < 0 ) {
		// we mirror sprites horizontally if walrus is facing left
		flags = flags + 64;
		temp = leftTile;
		leftTile = rightTile;
		rightTile = temp;
	}

	setsprite( 5+which*2+0, walrusX[ which ], walrusY[ which ], leftTile, flags );
	setsprite( 5+which*2+1, walrusX[ which ]+8, walrusY[ which ], rightTile, flags );
}

void MoveWalrus( unsigned char which ) {

	// do we need to turn around because we have reached the end of the screen?
	// right side?
	if( walrusX[ which ] + walrusDX[ which ] > (256 - 2*8) ) {
		walrusDX[ which ] = walrusDX[ which ] * (-1);
	}
	// left side?
	if( walrusX[ which ] + walrusDX[ which ] < 0 ) {
		walrusDX[ which ] = walrusDX[ which ] * (-1);
	}
	
	// move walrus
	walrusX[ which ] = walrusX[ which ] + walrusDX[ which ];
}

void InitializeGame() {
	unsigned int i;
	
	// walruses
	for( i = 0; i < MAX_WALRUSES; i++ ) {
		walrusVisible[i] = 1;
		// can't spawn walruses too far to the right; they must be on screen
		walrusX[i] = random() % (256 - 2*8);
		// can't spawn walruses too high, or they pick up bucket right away
		// can't spawn walruses too low, or bear can get hit right as the game begins
		walrusY[i] = (random() % (224 - 6*8)) + 3*8; 
		// some walruses are faster!
		walrusDX[ i ] = MIN_WALRUS_SPEED + random() % 2;
		if( i%2 == 0 ) 
			walrusDX[ i ] = walrusDX[ i ] * (-1);
	}

	// repeat our palette a few times to fill all palette slots
	for( i = 0; i < 8; i++ ) {
		memcpy(pal+32*i, paletteInfo, 32);
	}
	
	// load tiles
	settiles(0, tilesInfo, MAX_TILES*32);
} 

void RenderBear( unsigned int x, unsigned int y, unsigned char direction, unsigned long frame ) {

	unsigned int flags = 0x31;
	unsigned char temp;
	unsigned char topLeftTile = 0, topRightTile = 1, bottomLeftTile, bottomRightTile;
	
	if( frame % 40 < 10 ) {
		bottomLeftTile = 2, bottomRightTile = 3;
	} else if( frame % 40 < 20 ) {
		bottomLeftTile = 4; bottomRightTile = 5;
	} else if( frame % 40 < 30 ) {
		bottomLeftTile = 6;	bottomRightTile = 7;
	} else {
		bottomLeftTile = 4;	bottomRightTile = 5;
	}
	
	if( direction == 1 ) {
		// we mirror sprites horizontally based on direction
		flags = flags + 64;
		temp = topLeftTile;
		topLeftTile = topRightTile;
		topRightTile = temp;
		
		temp = bottomLeftTile;
		bottomLeftTile = bottomRightTile;
		bottomRightTile = temp;
	}
	
	setsprite(0, x, y, topLeftTile, flags);
	setsprite(1, x+8, y, topRightTile, flags);
	setsprite(2, x, y+8, bottomLeftTile, flags);
	setsprite(3, x+8, y+8, bottomRightTile, flags);
}

void RenderBucket( unsigned int x, unsigned int y ) {

	setsprite(4, x, y, 8, 0x31);
}

void HideWalrus( unsigned char which ) {
	walrusVisible[which] = 0;
}

int game() {
  snesc_init();
  char author[64] = "Bucket by Sebastian Mihai, 2012";
  unsigned long currentFrame = 0;
  int i, j, x=120, y=224 - 2*8;
  int bucketX = 256-8, bucketY = 0;
  unsigned char bearDirection = 0;
  unsigned int r;
  unsigned char bucketState = BUCKET_ON_GROUND;
  unsigned char atLeastOneWalrusLeft;

  InitializeGame();
  
  // generate snow on the ground procedurally
  for( j = 0; j < 28; j++ )
	for( i = 0; i < 32; i++ ) {
		r = (random() * i * j ) % 100;
		if( r > 0 && r < 3 )		
			SetBackgroundTile( i, j, 12 ); // first snow tile
		else if( r > 3 && r < 6 )
			SetBackgroundTile( i, j, 13 ); // second snow tile
		else
			SetBackgroundTile( i, j, 11 ); // empty tile
	}

	
  //writenum(lives, 8, blockmap, 0x136, 0x426);
  writestring("TEST!!!!!", blockmap, 0x248, 0x3F6);

  setmap(0, (unsigned char*)blockmap);
  setpalette((unsigned char*)pal);
  
  enablescreen();

  while(1) {
	resettimer();
	
	atLeastOneWalrusLeft = 0;
	currentFrame++;
	
	if ( ((getjoystatus(0) & RIGHT_BUTTON) != 0) && (x+BEAR_SPEED<256-2*8) ) {
		x += BEAR_SPEED;
		random();
		bearDirection = 0;
	} else if ( ((getjoystatus(0) & LEFT_BUTTON) != 0) && (x>BEAR_SPEED) ) {
		x -= BEAR_SPEED;
		random();
		bearDirection = 1;
	}
	
	if ( ((getjoystatus(0) & DOWN_BUTTON) != 0) && (y+BEAR_SPEED<224-2*8) ) {
		y += BEAR_SPEED;
		random();
	} else if ( ((getjoystatus(0) & UP_BUTTON) != 0) && (y>BEAR_SPEED) ) {
		y -= BEAR_SPEED;
		random();
	}
	
	for( i = 0; i < MAX_WALRUSES; i++ ) {
		MoveWalrus( i );
	}
	
	// is the bear picking up the bucket?
	if ( (getjoystatus(0) & A_BUTTON) != 0 ) {
	
		if(
			(
				( x < bucketX && bucketX - x < 16 ) ||
				( bucketX <= x && x - bucketX < 8 )
			)
			&&
			(
				( y < bucketY && bucketY - y < 16 ) ||
				( bucketY <= y && y - bucketY < 8 )
			)
		  ) {
				bucketState = BUCKET_HELD;
		  }
	
	}
	
	// is the bear dropping the bucket?
	if ( (getjoystatus(0) & B_BUTTON) != 0 ) {
		bucketState = BUCKET_ON_GROUND;
	}
	
	// the bucket sprite pretty much follows the bear's sprites if he's holding the bucket
	if( bucketState == BUCKET_HELD ) {
		if( bearDirection == 0 ) {
			bucketX = x+8; bucketY = y+4;
		} else {
			bucketX = x; bucketY = y+4;
		}
	}

	// walrus collision detection
	for( i = 0; i < MAX_WALRUSES; i++ ) {
		if( walrusVisible[i] == 0 ) continue;
		
		atLeastOneWalrusLeft = 1;
		// did this walrus collide with the bucket?
		if( bucketState == BUCKET_ON_GROUND ) {
			if(
				(
					( walrusX[i] < bucketX && bucketX - walrusX[i] < 16 ) ||
					( bucketX <= walrusX[i] && walrusX[i] - bucketX < 8 )
				)
				&&
				(
					( walrusY[i] < bucketY && bucketY - walrusY[i] < 8 ) ||
					( bucketY <= walrusY[i] && walrusY[i] - bucketY < 8 )
				)
			  ) {
					HideWalrus(i);
					
					// spawn new bucket in one of the four corners
					if( random() % 100 == 50 )
						bucketY = 0;
					else 
						bucketY = 224 - 8;
					
					if( random() % 100 < 50 )
						bucketX = 0;
					else 
						bucketX = 256-8;
					
					// this walrus is done, so move on to the next one
					continue;
						
			  }
		}
		
		// did this walrus collide with the bear?
		if(
			(
				( walrusX[i] < x && x - walrusX[i] < 14 ) ||
				( x <= walrusX[i] && walrusX[i] - x < 13 )
			)
			&&
			(
				( walrusY[i] < y && y - walrusY[i] < 6 ) ||
				( y <= walrusY[i] && walrusY[i] - y < 14 )
			)
		  ) {
				// game over
				delay(75); 
				return 0;
		  }
	}
	
	clearjoy(0);

	// RENDER THINGS
	RenderBear( x, y, bearDirection, currentFrame );
	RenderBucket( bucketX, bucketY );
	for( i = 0; i < MAX_WALRUSES; i++ ) {
		RenderWalrus( i );
	}
	
	if( atLeastOneWalrusLeft == 0 ) {
		while(1) {}
	}
	
	sync(1);
  }
}

int main() {
	while(1) {
		game();
	}
	
	return 0;
}
